@@ -27,18 +27,21 @@
 #include <QUndoStack>
 #include <QVariant>
 #include <QIcon>
+#include <QSet>
 
 #include "qgis_sip.h"
 #include "qgserror.h"
 #include "qgsobjectcustomproperties.h"
 #include "qgsrectangle.h"
 #include "qgscoordinatereferencesystem.h"
-#include "qgsrendercontext.h"
 #include "qgsmaplayerdependency.h"
 #include "qgslayermetadata.h"
+#include "qgsmaplayerserverproperties.h"
 #include "qgsmaplayerstyle.h"
 #include "qgsreadwritecontext.h"
 #include "qgsdataprovider.h"
+#include "qgis.h"
+#include "qgslogger.h"
 
 class QgsAbstract3DRenderer;
 class QgsDataProvider;
@@ -48,10 +51,12 @@ class QgsMapLayerStyleManager;
 class QgsProject;
 class QgsStyleEntityVisitorInterface;
 class QgsMapLayerTemporalProperties;
+class QgsMapLayerElevationProperties;
 
 class QDomDocument;
 class QKeyEvent;
 class QPainter;
+class QgsRenderContext;
 
 /*
  * Constants used to describe copy-paste MIME types
@@ -61,21 +66,7 @@ class QPainter;
 
 /**
  * \ingroup core
- * Types of layers that can be added to a map
- * \since QGIS 3.8
- */
-enum class QgsMapLayerType SIP_MONKEYPATCH_SCOPEENUM_UNNEST( QgsMapLayer, LayerType ) : int
-  {
-  VectorLayer,
-  RasterLayer,
-  PluginLayer,
-  MeshLayer,      //!< Added in 3.2
-  VectorTileLayer //!< Added in 3.14
-};
-
-/**
- * \ingroup core
- * Base class for all map layer types.
+ * \brief Base class for all map layer types.
  * This is the base class for all map layer types (vector, raster).
  */
 class CORE_EXPORT QgsMapLayer : public QObject
@@ -86,10 +77,13 @@ class CORE_EXPORT QgsMapLayer : public QObject
     Q_PROPERTY( int autoRefreshInterval READ autoRefreshInterval WRITE setAutoRefreshInterval NOTIFY autoRefreshIntervalChanged )
     Q_PROPERTY( QgsLayerMetadata metadata READ metadata WRITE setMetadata NOTIFY metadataChanged )
     Q_PROPERTY( QgsCoordinateReferenceSystem crs READ crs WRITE setCrs NOTIFY crsChanged )
+    Q_PROPERTY( QgsMapLayerType type READ type CONSTANT )
+    Q_PROPERTY( bool isValid READ isValid NOTIFY isValidChanged )
+    Q_PROPERTY( double opacity READ opacity WRITE setOpacity NOTIFY opacityChanged )
 
 #ifdef SIP_RUN
     SIP_CONVERT_TO_SUBCLASS_CODE
-    QgsMapLayer * layer = qobject_cast<QgsMapLayer *>( sipCpp );
+    QgsMapLayer *layer = qobject_cast<QgsMapLayer *>( sipCpp );
 
     sipType = 0;
 
@@ -112,6 +106,15 @@ class CORE_EXPORT QgsMapLayer : public QObject
         case QgsMapLayerType::VectorTileLayer:
           sipType = sipType_QgsVectorTileLayer;
           break;
+        case QgsMapLayerType::AnnotationLayer:
+          sipType = sipType_QgsAnnotationLayer;
+          break;
+        case QgsMapLayerType::PointCloudLayer:
+          sipType = sipType_QgsPointCloudLayer;
+          break;
+        case QgsMapLayerType::GroupLayer:
+          sipType = sipType_QgsGroupLayer;
+          break;
         default:
           sipType = nullptr;
           break;
@@ -142,6 +145,7 @@ class CORE_EXPORT QgsMapLayer : public QObject
       Identifiable = 1 << 0, //!< If the layer is identifiable using the identify map tool and as a WMS layer.
       Removable = 1 << 1,    //!< If the layer can be removed from the project. The layer will not be removable from the legend menu entry but can still be removed with an API call.
       Searchable = 1 << 2,   //!< Only for vector-layer, determines if the layer is used in the 'search all layers' locator.
+      Private = 1 << 3,       //!< Determines if the layer is meant to be exposed to the GUI, i.e. visible in the layer legend tree.
     };
     Q_ENUM( LayerFlag )
     Q_DECLARE_FLAGS( LayerFlags, LayerFlag )
@@ -167,9 +171,12 @@ class CORE_EXPORT QgsMapLayer : public QObject
       CustomProperties   = 1 << 11, //!< Custom properties (by plugins for instance)
       GeometryOptions    = 1 << 12, //!< Geometry validation configuration
       Relations          = 1 << 13, //!< Relations
-      Temporal           = 1 << 14, //!< Temporal properties
+      Temporal           = 1 << 14, //!< Temporal properties (since QGIS 3.14)
+      Legend             = 1 << 15, //!< Legend settings (since QGIS 3.16)
+      Elevation          = 1 << 16, //!< Elevation settings (since QGIS 3.18)
+      Notes              = 1 << 17, //!< Layer user notes (since QGIS 3.20)
       AllStyleCategories = LayerConfiguration | Symbology | Symbology3D | Labeling | Fields | Forms | Actions |
-                           MapTips | Diagrams | AttributeTable | Rendering | CustomProperties | GeometryOptions | Relations | Temporal,
+                           MapTips | Diagrams | AttributeTable | Rendering | CustomProperties | GeometryOptions | Relations | Temporal | Legend | Elevation | Notes,
     };
     Q_ENUM( StyleCategory )
     Q_DECLARE_FLAGS( StyleCategories, StyleCategory )
@@ -208,19 +215,36 @@ class CORE_EXPORT QgsMapLayer : public QObject
      * \note Flags are options specified by the user used for the UI but are not preventing any API call.
      * For instance, even if the Removable flag is not set, the layer can still be removed with the API
      * but the action will not be listed in the legend menu.
+     *
+     * \see properties()
+     *
      * \since QGIS 3.4
      */
     QgsMapLayer::LayerFlags flags() const;
 
     /**
      * Returns the flags for this layer.
-      \note Flags are options specified by the user used for the UI but are not preventing any API call.
+     * \note Flags are options specified by the user used for the UI but are not preventing any API call.
      * For instance, even if the Removable flag is not set, the layer can still be removed with the API
      * but the action will not be listed in the legend menu.
+     *
+     * \see properties()
+     *
      * \since QGIS 3.4
      */
     void setFlags( QgsMapLayer::LayerFlags flags );
 
+    /**
+     * Returns the map layer properties of this layer.
+     *
+     * \note properties() differ from flags() in that flags() are user settable, and reflect options that
+     * users can enable for map layers. In contrast properties() are reflections of inherent capabilities
+     * for the layer, which cannot be directly changed by users.
+     *
+     * \since QGIS 3.22
+     */
+    virtual Qgis::MapLayerProperties properties() const;
+
     /**
      * Returns the extension of a Property.
      * \returns The extension
@@ -247,7 +271,7 @@ class CORE_EXPORT QgsMapLayer : public QObject
     /**
      * Returns the layer's data provider, it may be NULLPTR.
      */
-    virtual QgsDataProvider *dataProvider();
+    Q_INVOKABLE virtual QgsDataProvider *dataProvider();
 
     /**
      * Returns the layer's data provider in a const-correct manner, it may be NULLPTR.
@@ -395,59 +419,81 @@ class CORE_EXPORT QgsMapLayer : public QObject
 
     /* Layer metadataUrl information */
 
+    /**
+     * Returns QGIS Server Properties for the map layer
+     * \since QGIS 3.22
+     */
+    QgsMapLayerServerProperties *serverProperties() { return mServerProperties.get(); };
+
+    /**
+     * Returns QGIS Server Properties const for the map layer
+     * \since QGIS 3.22
+     */
+    const QgsMapLayerServerProperties *serverProperties() const { return mServerProperties.get(); } SIP_SKIP;
+
     /**
      * Sets the metadata URL of the layer
      *  used by QGIS Server in GetCapabilities request.
      *  MetadataUrl is a a link to the detailed, standardized metadata about the data.
-     * \returns the layer metadata URL
-     * \see metadataUrl()
+     *  Since QGIS 3.22, it edits the first metadata URL link.
+     * \see serverProperties()
+     * \deprecated since QGIS 3.22
      */
-    void setMetadataUrl( const QString &metaUrl ) { mMetadataUrl = metaUrl; }
+    Q_DECL_DEPRECATED void setMetadataUrl( const QString &metaUrl ) SIP_DEPRECATED;
 
     /**
      * Returns the metadata URL of the layer
      *  used by QGIS Server in GetCapabilities request.
      *  MetadataUrl is a a link to the detailed, standardized metadata about the data.
+     * Since QGIS 3.22, it returns the first metadata URL link.
      * \returns the layer metadata URL
-     * \see setMetadataUrl()
+     * \see serverProperties()
+     * \deprecated since QGIS 3.22
      */
-    QString metadataUrl() const { return mMetadataUrl; }
+    Q_DECL_DEPRECATED QString metadataUrl() const SIP_DEPRECATED;
 
     /**
      * Set the metadata type of the layer
      *  used by QGIS Server in GetCapabilities request
      *  MetadataUrlType indicates the standard to which the metadata complies.
+     *  Since QGIS 3.22, it edits the first metadata URL type.
      * \returns the layer metadata type
-     * \see metadataUrlType()
+     * \see serverProperties()
+     * \deprecated since QGIS 3.22
      */
-    void setMetadataUrlType( const QString &metaUrlType ) { mMetadataUrlType = metaUrlType; }
+    Q_DECL_DEPRECATED void setMetadataUrlType( const QString &metaUrlType ) SIP_DEPRECATED;
 
     /**
      * Returns the metadata type of the layer
      *  used by QGIS Server in GetCapabilities request.
      *  MetadataUrlType indicates the standard to which the metadata complies.
+     * Since QGIS 3.22, it returns the first metadata URL type.
      * \returns the layer metadata type
-     * \see setMetadataUrlType()
+     * \see serverProperties()
+     * \deprecated since QGIS 3.22
      */
-    QString metadataUrlType() const { return mMetadataUrlType; }
+    Q_DECL_DEPRECATED QString metadataUrlType() const SIP_DEPRECATED;
 
     /**
      * Sets the metadata format of the layer
      *  used by QGIS Server in GetCapabilities request.
      *  MetadataUrlType indicates how the metadata is structured.
-     * \returns the layer metadata format
-     * \see metadataUrlFormat()
+     *  Since QGIS 3.22, it edits the first metadata URL format.
+     * \see serverProperties()
+     * \deprecated since QGIS 3.22
      */
-    void setMetadataUrlFormat( const QString &metaUrlFormat ) { mMetadataUrlFormat = metaUrlFormat; }
+    Q_DECL_DEPRECATED void setMetadataUrlFormat( const QString &metaUrlFormat ) SIP_DEPRECATED;
 
     /**
      * Returns the metadata format of the layer
      *  used by QGIS Server in GetCapabilities request.
      *  MetadataUrlType indicates how the metadata is structured.
+     * Since QGIS 3.22, it returns the first metadata URL format.
      * \returns the layer metadata format
-     * \see setMetadataUrlFormat()
+     * \see serverProperties()
+     * \deprecated since QGIS 3.22
      */
-    QString metadataUrlFormat() const { return mMetadataUrlFormat; }
+    Q_DECL_DEPRECATED QString metadataUrlFormat() const SIP_DEPRECATED;
 
     /**
      * Set the blending mode used for rendering a layer.
@@ -462,13 +508,33 @@ class CORE_EXPORT QgsMapLayer : public QObject
     */
     QPainter::CompositionMode blendMode() const;
 
+    /**
+     * Sets the \a opacity for the layer, where \a opacity is a value between 0 (totally transparent)
+     * and 1.0 (fully opaque).
+     * \see opacity()
+     * \see opacityChanged()
+     * \note Prior to QGIS 3.18, this method was available for vector layers only
+     * \since QGIS 3.18
+     */
+    virtual void setOpacity( double opacity );
+
+    /**
+     * Returns the opacity for the layer, where opacity is a value between 0 (totally transparent)
+     * and 1.0 (fully opaque).
+     * \see setOpacity()
+     * \see opacityChanged()
+     * \note Prior to QGIS 3.18, this method was available for vector layers only
+     * \since QGIS 3.18
+     */
+    virtual double opacity() const;
+
     //! Returns if this layer is read only.
     bool readOnly() const { return isReadOnly(); }
 
     /**
      * Synchronises with changes in the datasource
-        */
-    virtual void reload() {}
+     */
+    Q_INVOKABLE virtual void reload() {}
 
     /**
      * Returns new instance of QgsMapLayerRenderer that will be used for rendering of given context
@@ -479,6 +545,16 @@ class CORE_EXPORT QgsMapLayer : public QObject
     //! Returns the extent of the layer.
     virtual QgsRectangle extent() const;
 
+    /**
+     * Returns the WGS84 extent (EPSG:4326) of the layer according to
+     * ReadFlag::FlagTrustLayerMetadata. If that flag is activated, then the
+     * WGS84 extent read in the qgs project is returned. Otherwise, the actual
+     * WGS84 extent is returned.
+     * \param forceRecalculate True to return the current WGS84 extent whatever the read flags
+     * \since QGIS 3.20
+     */
+    QgsRectangle wgs84Extent( bool forceRecalculate = false ) const;
+
     /**
      * Returns the status of the layer. An invalid layer is one which has a bad datasource
      * or other problem. Child classes set this flag when initialized.
@@ -520,9 +596,24 @@ class CORE_EXPORT QgsMapLayer : public QObject
     */
     virtual void setSubLayerVisibility( const QString &name, bool visible );
 
+    /**
+     * Returns whether the layer supports editing or not.
+     * \returns FALSE if the layer is read only or the data provider has no editing capabilities.
+     * \note default implementation returns FALSE.
+     * \since QGIS 3.22 in the base class QgsMapLayer.
+     */
+    virtual bool supportsEditing() const;
+
     //! Returns TRUE if the layer can be edited.
     virtual bool isEditable() const;
 
+    /**
+     * Returns TRUE if the layer has been modified since last commit/save.
+     * \note default implementation returns FALSE.
+     * \since QGIS 3.22 in the base class QgsMapLayer.
+     */
+    virtual bool isModified() const;
+
     /**
      * Returns TRUE if the layer is considered a spatial layer, ie it has some form of geometry associated with it.
      * \since QGIS 2.16
@@ -532,8 +623,9 @@ class CORE_EXPORT QgsMapLayer : public QObject
     /**
      * Returns TRUE if the layer is considered a temporary layer.
      *
-     * These include memory-only layers such as those created by the "memory" data provider, or layers
-     * stored inside a local temporary folder (such as the "/tmp" folder on Linux).
+     * These include memory-only layers such as those created by the "memory" data provider, layers
+     * stored inside a local temporary folder (such as the "/tmp" folder on Linux)
+     * or layer with temporary data (as temporary mesh layer dataset)
      *
      * \since QGIS 3.10.1
      */
@@ -546,6 +638,8 @@ class CORE_EXPORT QgsMapLayer : public QObject
     enum ReadFlag
     {
       FlagDontResolveLayers = 1 << 0, //!< Don't resolve layer paths or create data providers for layers.
+      FlagTrustLayerMetadata = 1 << 1, //!< Trust layer metadata. Improves layer load time by skipping expensive checks like primary key unicity, geometry type and srid and by using estimated metadata on layer load. Since QGIS 3.16
+      FlagReadExtentFromXml = 1 << 2, //!< Read extent from xml and skip get extent from provider.
     };
     Q_DECLARE_FLAGS( ReadFlags, ReadFlag )
 
@@ -626,6 +720,177 @@ class CORE_EXPORT QgsMapLayer : public QObject
      */
     const QgsObjectCustomProperties &customProperties() const;
 
+#ifndef SIP_RUN
+
+    /**
+     * Returns the property value for a property based on an enum.
+     * This forces the output to be a valid and existing entry of the enum.
+     * Hence if the property value is incorrect, the given default value is returned.
+     * This tries first with property as a string (as the enum) and then as an integer value.
+     * \note The enum needs to be declared with Q_ENUM, and flags with Q_FLAG (not Q_FLAGS).
+     * \see setCustomEnumProperty
+     * \see customFlagProperty
+     * \since QGIS 3.22
+     */
+    template <class T>
+    T customEnumProperty( const QString &key, const T &defaultValue )
+    {
+      const QMetaEnum metaEnum = QMetaEnum::fromType<T>();
+      Q_ASSERT( metaEnum.isValid() );
+      if ( !metaEnum.isValid() )
+      {
+        QgsDebugMsg( QStringLiteral( "Invalid metaenum. Enum probably misses Q_ENUM or Q_FLAG declaration." ) );
+      }
+
+      T v;
+      bool ok = false;
+
+      if ( metaEnum.isValid() )
+      {
+        // read as string
+        QByteArray ba = customProperty( key, metaEnum.valueToKey( static_cast<int>( defaultValue ) ) ).toString().toUtf8();
+        const char *vs = ba.data();
+        v = static_cast<T>( metaEnum.keyToValue( vs, &ok ) );
+        if ( ok )
+          return v;
+      }
+
+      // if failed, try to read as int (old behavior)
+      // this code shall be removed later
+      // then the method could be marked as const
+      v = static_cast<T>( customProperty( key, static_cast<int>( defaultValue ) ).toInt( &ok ) );
+      if ( metaEnum.isValid() )
+      {
+        if ( !ok || !metaEnum.valueToKey( static_cast<int>( v ) ) )
+        {
+          v = defaultValue;
+        }
+        else
+        {
+          // found property as an integer
+          // convert the property to the new form (string)
+          setCustomEnumProperty( key, v );
+        }
+      }
+
+      return v;
+    }
+
+    /**
+     * Set the value of a property based on an enum.
+     * The property will be saved as string.
+     * \note The enum needs to be declared with Q_ENUM, and flags with Q_FLAG (not Q_FLAGS).
+     * \see customEnumProperty
+     * \see setCustomFlagProperty
+     * \since QGIS 3.22
+     */
+    template <class T>
+    void setCustomEnumProperty( const QString &key, const T &value )
+    {
+      const QMetaEnum metaEnum = QMetaEnum::fromType<T>();
+      Q_ASSERT( metaEnum.isValid() );
+      if ( metaEnum.isValid() )
+      {
+        setCustomProperty( key, metaEnum.valueToKey( static_cast<int>( value ) ) );
+      }
+      else
+      {
+        QgsDebugMsg( QStringLiteral( "Invalid metaenum. Enum probably misses Q_ENUM or Q_FLAG declaration." ) );
+      }
+    }
+
+    /**
+     * Returns the property value for a property based on a flag.
+     * This forces the output to be a valid and existing entry of the flag.
+     * Hence if the property value is incorrect, the given default value is returned.
+     * This tries first with property as a string (using a byte array) and then as an integer value.
+     * \note The flag needs to be declared with Q_FLAG (not Q_FLAGS).
+     * \note for Python bindings, a custom implementation is achieved in Python directly.
+     * \see setCustomFlagProperty
+     * \see customEnumProperty
+     * \since QGIS 3.22
+     */
+    template <class T>
+    T customFlagProperty( const QString &key, const T &defaultValue )
+    {
+      const QMetaEnum metaEnum = QMetaEnum::fromType<T>();
+      Q_ASSERT( metaEnum.isValid() );
+      if ( !metaEnum.isValid() )
+      {
+        QgsDebugMsg( QStringLiteral( "Invalid metaenum. Enum probably misses Q_ENUM or Q_FLAG declaration." ) );
+      }
+
+      T v = defaultValue;
+      bool ok = false;
+
+      if ( metaEnum.isValid() )
+      {
+        // read as string
+        QByteArray ba = customProperty( key, metaEnum.valueToKeys( defaultValue ) ).toString().toUtf8();
+        const char *vs = ba.data();
+        v = static_cast<T>( metaEnum.keysToValue( vs, &ok ) );
+      }
+      if ( !ok )
+      {
+        // if failed, try to read as int
+        const int intValue = customProperty( key, static_cast<int>( defaultValue ) ).toInt( &ok );
+        if ( metaEnum.isValid() )
+        {
+          if ( ok )
+          {
+            // check that the int value does correspond to a flag
+            // see https://stackoverflow.com/a/68495949/1548052
+            const QByteArray keys = metaEnum.valueToKeys( intValue );
+            const int intValueCheck = metaEnum.keysToValue( keys );
+            if ( intValue != intValueCheck )
+            {
+              v = defaultValue;
+            }
+            else
+            {
+              // found property as an integer
+              v = T( intValue );
+              // convert the property to the new form (string)
+              // this code could be removed
+              // then the method could be marked as const
+              setCustomFlagProperty( key, v );
+            }
+          }
+          else
+          {
+            v = defaultValue;
+          }
+        }
+      }
+
+      return v;
+    }
+
+    /**
+     * Set the value of a property based on a flag.
+     * The property will be saved as string.
+     * \note The flag needs to be declared with Q_FLAG (not Q_FLAGS).
+     * \see customFlagProperty
+     * \see customEnumProperty
+     * \since QGIS 3.22
+     */
+    template <class T>
+    void setCustomFlagProperty( const QString &key, const T &value )
+    {
+      const QMetaEnum metaEnum = QMetaEnum::fromType<T>();
+      Q_ASSERT( metaEnum.isValid() );
+      if ( metaEnum.isValid() )
+      {
+        setCustomProperty( key, metaEnum.valueToKeys( value ) );
+      }
+      else
+      {
+        QgsDebugMsg( QStringLiteral( "Invalid metaenum. Enum probably misses Q_ENUM or Q_FLAG declaration." ) );
+      }
+    }
+#endif
+
+
     /**
      * Remove a custom property from layer. Properties are stored in a map and saved in project file.
      * \see setCustomProperty()
@@ -641,7 +906,7 @@ class CORE_EXPORT QgsMapLayer : public QObject
 
     /**
      * Returns the layer's spatial reference system.
-    \since QGIS 1.4
+     * \since QGIS 1.4
      */
     QgsCoordinateReferenceSystem crs() const;
 
@@ -732,7 +997,7 @@ class CORE_EXPORT QgsMapLayer : public QObject
      * \returns a QString with any status messages
      * \since QGIS 3.0
      */
-    QString loadDefaultMetadata( bool &resultFlag );
+    virtual QString loadDefaultMetadata( bool &resultFlag );
 
     /**
      * Retrieve a named metadata for this layer from a sqlite database.
@@ -909,7 +1174,7 @@ class CORE_EXPORT QgsMapLayer : public QObject
                             QgsReadWriteContext &context, StyleCategories categories = AllStyleCategories );
 
     /**
-     * Write the style for the layer into the docment provided.
+     * Write the style for the layer into the document provided.
      *  \param node the node that will have the style element added to it.
      *  \param doc the document that will have the QDomNode added.
      *  \param errorMessage reference to string that will be updated with any error messages
@@ -941,7 +1206,21 @@ class CORE_EXPORT QgsMapLayer : public QObject
      * Updates the data source of the layer. The layer's renderer and legend will be preserved only
      * if the geometry type of the new data source matches the current geometry type of the layer.
      *
-     * Subclasses should override this method: default implementation does nothing.
+     * This method was defined in QgsVectorLayer since 2.10 and was marked as deprecated since 3.2
+     *
+     * \param dataSource new layer data source
+     * \param baseName base name of the layer
+     * \param provider provider string
+     * \param loadDefaultStyleFlag set to TRUE to reset the layer's style to the default for the
+     * data source
+     * \see dataSourceChanged()
+     * \since QGIS 3.20
+     */
+    void setDataSource( const QString &dataSource, const QString &baseName, const QString &provider, bool loadDefaultStyleFlag = false );
+
+    /**
+     * Updates the data source of the layer. The layer's renderer and legend will be preserved only
+     * if the geometry type of the new data source matches the current geometry type of the layer.
      *
      * \param dataSource new layer data source
      * \param baseName base name of the layer
@@ -952,7 +1231,23 @@ class CORE_EXPORT QgsMapLayer : public QObject
      * \see dataSourceChanged()
      * \since QGIS 3.6
      */
-    virtual void setDataSource( const QString &dataSource, const QString &baseName, const QString &provider, const QgsDataProvider::ProviderOptions &options, bool loadDefaultStyleFlag = false );
+    void setDataSource( const QString &dataSource, const QString &baseName, const QString &provider, const QgsDataProvider::ProviderOptions &options, bool loadDefaultStyleFlag = false );
+
+    /**
+     * Updates the data source of the layer. The layer's renderer and legend will be preserved only
+     * if the geometry type of the new data source matches the current geometry type of the layer.
+     *
+     * Subclasses should override setDataSourcePrivate: default implementation does nothing.
+     *
+     * \param dataSource new layer data source
+     * \param baseName base name of the layer
+     * \param provider provider string
+     * \param options provider options
+     * \param flags provider read flags which control dataprovider construction like FlagTrustDataSource, FlagLoadDefaultStyle, etc
+     * \see dataSourceChanged()
+     * \since QGIS 3.20
+     */
+    void setDataSource( const QString &dataSource, const QString &baseName, const QString &provider, const QgsDataProvider::ProviderOptions &options, QgsDataProvider::ReadFlags flags );
 
     /**
      * Returns the provider type (provider key) for this layer
@@ -1193,6 +1488,27 @@ class CORE_EXPORT QgsMapLayer : public QObject
      */
     virtual QgsMapLayerTemporalProperties *temporalProperties() { return nullptr; }
 
+    /**
+     * Returns the layer's elevation properties. This may be NULLPTR, depending on the layer type.
+     *
+     * \since QGIS 3.18
+     */
+    virtual QgsMapLayerElevationProperties *elevationProperties() { return nullptr; }
+
+    /**
+     * Returns path to the placeholder image or an empty string if a generated legend is shown
+     * \return placholder image path
+     * \since QGIS 3.22
+     */
+    QString legendPlaceholderImage() const { return mLegendPlaceholderImage;}
+
+    /**
+     * Set placeholder image for legend. If the string is empty, a generated legend will be shown.
+     * \param imgPath file path to the placeholder image
+     * \since QGIS 3.22
+     */
+    void setLegendPlaceholderImage( const QString &imgPath ) { mLegendPlaceholderImage = imgPath; }
+
   public slots:
 
     /**
@@ -1236,6 +1552,14 @@ class CORE_EXPORT QgsMapLayer : public QObject
      */
     void triggerRepaint( bool deferredUpdate = false );
 
+    /**
+     * Will advise any 3D maps that this layer requires to be updated in the scene.
+     * Will emit a request3DUpdate() signal.
+     *
+     * \since QGIS 3.18
+     */
+    void trigger3DUpdate();
+
     /**
      * Triggers an emission of the styleChanged() signal.
      * \since QGIS 2.16
@@ -1283,6 +1607,14 @@ class CORE_EXPORT QgsMapLayer : public QObject
     % End
 #endif
 
+    /**
+     * Returns the parent project if this map layer is added to a project.
+     * Otherwise returns NULLPTR
+     *
+     * \since QGIS 3.18
+     */
+    QgsProject *project() const;
+
   signals:
 
     /**
@@ -1323,6 +1655,16 @@ class CORE_EXPORT QgsMapLayer : public QObject
     //! Signal emitted when the blend mode is changed, through QgsMapLayer::setBlendMode()
     void blendModeChanged( QPainter::CompositionMode blendMode );
 
+    /**
+     * Emitted when the layer's opacity is changed, where \a opacity is a value between 0 (transparent)
+     * and 1 (opaque).
+     * \see setOpacity()
+     * \see opacity()
+     * \note Prior to QGIS 3.18, this signal was available for vector layers only
+     * \since QGIS 3.18
+     */
+    void opacityChanged( double opacity );
+
     /**
      * Signal emitted when renderer is changed.
      * \see styleChanged()
@@ -1333,6 +1675,10 @@ class CORE_EXPORT QgsMapLayer : public QObject
      * Signal emitted whenever a change affects the layer's style. Ie this may be triggered
      * by renderer changes, label style changes, or other style changes such as blend
      * mode or layer opacity changes.
+     *
+     * \warning This signal should never be manually emitted. Instead call the emitStyleChanged() method
+     * to ensure that the signal is only emitted when appropriate.
+     *
      * \see rendererChanged()
      * \since QGIS 2.16
     */
@@ -1350,6 +1696,13 @@ class CORE_EXPORT QgsMapLayer : public QObject
      */
     void renderer3DChanged();
 
+    /**
+     * Signal emitted when a layer requires an update in any 3D maps.
+     *
+     * \since QGIS 3.18
+     */
+    void request3DUpdate();
+
     /**
      * Emitted whenever the configuration is changed. The project listens to this signal
      * to be marked as dirty.
@@ -1408,10 +1761,58 @@ class CORE_EXPORT QgsMapLayer : public QObject
      */
     void styleLoaded( QgsMapLayer::StyleCategories categories );
 
+    /**
+     * Emitted when the validity of this layer changed.
+     *
+     * \since QGIS 3.16
+     */
+    void isValidChanged();
+
+    /**
+     * Emitted when a custom property of the layer has been changed or removed.
+     *
+     * \since QGIS 3.18
+     */
+    void customPropertyChanged( const QString &key );
+
+    /**
+     * Emitted when editing on this layer has started.
+     * \since QGIS 3.22 in the QgsMapLayer base class
+     */
+    void editingStarted();
+
+    /**
+     * Emitted when edited changes have been successfully written to the data provider.
+     * \since QGIS 3.22 in the QgsMapLayer base class
+     */
+    void editingStopped();
+
+    /**
+     * Emitted when modifications has been done on layer
+     * \since QGIS 3.22 in the QgsMapLayer base class
+     */
+    void layerModified();
 
   private slots:
 
-    void onNotifiedTriggerRepaint( const QString &message );
+    void onNotified( const QString &message );
+
+    /**
+     * Updates the data source of the layer. The layer's renderer and legend will be preserved only
+     * if the geometry type of the new data source matches the current geometry type of the layer.
+     *
+     * Called by setDataSource()
+     * Subclasses should override this method: default implementation does nothing.
+     *
+     * \param dataSource new layer data source
+     * \param baseName base name of the layer
+     * \param provider provider string
+     * \param options provider options
+     * \param flags provider read flags
+     * \see dataSourceChanged()
+     * \since QGIS 3.20
+     */
+    virtual void setDataSourcePrivate( const QString &dataSource, const QString &baseName, const QString &provider, const QgsDataProvider::ProviderOptions &options, QgsDataProvider::ReadFlags flags );
 
   protected:
 
@@ -1425,7 +1826,7 @@ class CORE_EXPORT QgsMapLayer : public QObject
     //! Sets the extent
     virtual void setExtent( const QgsRectangle &rect );
 
-    //! Sets whether layer is valid or not - should be used in constructor.
+    //! Sets whether layer is valid or not
     void setValid( bool valid );
 
     /**
@@ -1469,8 +1870,9 @@ class CORE_EXPORT QgsMapLayer : public QObject
 
     /**
      * Read custom properties from project file.
-      \param layerNode note to read from
-      \param keyStartsWith reads only properties starting with the specified string (or all if the string is empty)*/
+     * \param layerNode note to read from
+     * \param keyStartsWith reads only properties starting with the specified string (or all if the string is empty)
+    */
     void readCustomProperties( const QDomNode &layerNode, const QString &keyStartsWith = QString() );
 
     //! Write custom properties to project file.
@@ -1511,8 +1913,13 @@ class CORE_EXPORT QgsMapLayer : public QObject
     //! Sets error message
     void setError( const QgsError &error ) { mError = error;}
 
-    //! Extent of the layer
-    mutable QgsRectangle mExtent;
+    /**
+     * Invalidates the WGS84 extent. If FlagTrustLayerMetadata is enabled,
+     * the extent is not invalidated because we want to trust metadata whatever
+     * happens.
+     * \since QGIS 3.20
+     */
+    void invalidateWgs84Extent();
 
     //! Indicates if the layer is valid and can be drawn
     bool mValid = false;
@@ -1538,11 +1945,6 @@ class CORE_EXPORT QgsMapLayer : public QObject
     QString mAttribution;
     QString mAttributionUrl;
 
-    //! MetadataUrl of the layer
-    QString mMetadataUrl;
-    QString mMetadataUrlType;
-    QString mMetadataUrlFormat;
-
     //! WMS legend
     QString mLegendUrl;
     QString mLegendUrlFormat;
@@ -1555,7 +1957,7 @@ class CORE_EXPORT QgsMapLayer : public QObject
 
     /**
      * Checks whether a new set of dependencies will introduce a cycle
-     * this method is now deprecated and always return false, because circular dependencies are now correctly managed.
+     * this method is now deprecated and always return FALSE, because circular dependencies are now correctly managed.
      * \deprecated since QGIS 3.10
      */
     Q_DECL_DEPRECATED bool hasDependencyCycle( const QSet<QgsMapLayerDependency> & ) const {return false;}
@@ -1578,6 +1980,46 @@ class CORE_EXPORT QgsMapLayer : public QObject
      */
     bool mShouldValidateCrs = true;
 
+    /**
+     * Layer opacity.
+     *
+     * \since QGIS 3.18
+     */
+    double mLayerOpacity = 1.0;
+
+    /**
+     * If non-zero, the styleChanged signal should not be emitted.
+     *
+     * \since QGIS 3.20
+     */
+    int mBlockStyleChangedSignal = 0;
+
+#ifndef SIP_RUN
+
+    /**
+     * Returns a HTML fragment containing the layer's CRS metadata, for use
+     * in the htmlMetadata() method.
+     *
+     * \note Not available in Python bindings.
+     *
+     * \since QGIS 3.20
+     */
+    QString crsHtmlMetadata() const;
+#endif
+
+#ifndef SIP_RUN
+
+    /**
+     * Returns an HTML fragment containing general  metadata information, for use
+     * in the htmlMetadata() method.
+     *
+     * \note Not available in Python bindings.
+     *
+     * \since QGIS 3.22
+     */
+    QString generalHtmlMetadata() const;
+#endif
+
   private:
 
     virtual QString baseURI( PropertyType type ) const;
@@ -1587,6 +2029,9 @@ class CORE_EXPORT QgsMapLayer : public QObject
                                bool &resultFlag, StyleCategories categories = AllStyleCategories );
     bool loadNamedPropertyFromDatabase( const QString &db, const QString &uri, QString &xml, QgsMapLayer::PropertyType type );
 
+    // const method because extents are mutable
+    void updateExtent( const QgsRectangle &extent ) const;
+
     /**
      * This method returns TRUE by default but can be overwritten to specify
      * that a certain layer is writable.
@@ -1595,7 +2040,8 @@ class CORE_EXPORT QgsMapLayer : public QObject
 
     /**
      * Layer's spatial reference system.
-        private to make sure setCrs must be used and crsChanged() is emitted */
+     * private to make sure setCrs must be used and crsChanged() is emitted.
+    */
     QgsCoordinateReferenceSystem mCRS;
 
     //! Unique ID of this layer - used to refer to this layer in map layer registry
@@ -1621,7 +2067,12 @@ class CORE_EXPORT QgsMapLayer : public QObject
     //! A flag that tells us whether to use the above vars to restrict layer visibility
     bool mScaleBasedVisibility = false;
 
-    //! Collection of undoable operations for this layer. *
+    /**
+     * Stores information about server properties
+     */
+    std::unique_ptr< QgsMapLayerServerProperties > mServerProperties;
+
+    //! Collection of undoable operations for this layer.
     QUndoStack *mUndoStack = nullptr;
 
     QUndoStack *mUndoStackStyles = nullptr;
@@ -1643,6 +2094,12 @@ class CORE_EXPORT QgsMapLayer : public QObject
     //! Renderer for 3D views
     QgsAbstract3DRenderer *m3DRenderer = nullptr;
 
+    //! Extent of the layer
+    mutable QgsRectangle mExtent;
+
+    //! Extent of the layer in EPSG:4326
+    mutable QgsRectangle mWgs84Extent;
+
     /**
      * Stores the original XML properties of the layer when loaded from the project
      *
@@ -1652,6 +2109,12 @@ class CORE_EXPORT QgsMapLayer : public QObject
 
     //! To avoid firing multiple time repaintRequested signal on circular layer circular dependencies
     bool mRepaintRequestedFired = false;
+
+    //! Path to placeholder image for layer legend. If the string is empty, a generated legend is shown
+    QString mLegendPlaceholderImage;
+
+    friend class QgsVectorLayer;
+    friend class TestQgsMapLayer;
 };
 
 Q_DECLARE_METATYPE( QgsMapLayer * )
